//version: 1.6
#region Using directives

using System;
using System.Collections;
using System.Windows.Forms;
using System.Drawing;
using System.Reflection;
using System.Globalization;
using SysBindingFlags = System.Reflection.BindingFlags;

#endregion

namespace Reflection
{
    public class Object
    {

        // 1.1 added ValueAs properties
        // 1.2 added Field get and set methods
        //     added ValueAsUIntxx properties
        //     added GetXXX Reflection based methods
        //     merged official Door library code
        // 1.3 added catch for ambiguous match in SetProperty5/8
        //     added KeyType parameter to GetProperty3/4 and SetProperty5/8 to resolve ambiguous match
        //     change ObjectEx.Type property to ObjectEx.ObjectType
        //____________________________________________________________________________
        //
        // 1.0 changed name to Reflection library
        //     changed my extra methods to use global ot instead of a local o.GetType() for static use
        //     added bindingflags and ValueAsEnum
        // 1.1 rationalised GetPropertyX and SetPropertyX code
        //     added LastFieldInfo, LastMethodInfo, LastPropertyInfo
        //     added GetFieldByInfo, SetFieldByInfo, RunMethodByInfo
        //     added GetPropertyByInfo, GetPropertyByInfo2, SetPropertyByInfo, SetPropertyByInfo2
        //     System_NS changed to System_
        //     Event can now add any event as the desktop can create any delegate
        // 1.2 Added Convert to all get ValueAs because can't directly cast a boxed primitive
        // 1.3 Added AssemblyFullName to get the full name of any strong named assembly
        //     Changed all the Sytem_xxx properties to use AssemblyFullName so avoiding loading unneeded assemblies.
        // 1.4 Added BindingFlags and associated FlagXXX properties
        // 1.5 Removed references to System.Web, Xml etc left over from 1.3 change
        //     Added missing ", " omitted from the System_XXX properties added in 1.3
        // 1.6 Removed SetField2 and some SetProperties methods as the IDe can now pass strings to library Object parameters


        private double version = 1.6;

        // start of original Door library Object object code

        private object o;
        private Type ot;
        private BindingFlags bindingflags = SysBindingFlags.Public | SysBindingFlags.Instance | SysBindingFlags.Static | SysBindingFlags.NonPublic;
        private FieldInfo lastField;
        private MethodInfo lastMethod;
        private PropertyInfo lastProperty;


        public string System_Mscorlib // not really needed, used by default
        { get { return AssemblyFullName("mscorlib");  } }

        public string System_
        { get { return AssemblyFullName("System"); } }

        public string System_Data
        { get { return AssemblyFullName("System.Data"); } }

        public string System_Drawing
        { get { return AssemblyFullName("System.Drawing"); } }

        public string System_Web
        { get { return AssemblyFullName("System.Web"); } }

        public string System_Windows_Forms
        { get { return AssemblyFullName("System.Windows.Forms"); } }

        public string System_Xml
        { get { return AssemblyFullName("System.Xml"); } }


        // Constructor
        public Object()
        {
        }


        // Public Methods
        public string AssemblyFullName(string assembly)
        {
            #pragma warning disable 618
            Assembly ass = Assembly.LoadWithPartialName(assembly);
            #pragma warning restore 618
            return ", " + ass.FullName;
        }


        public void CreateNew(string Type)
        {
            ot = System.Type.GetType(Type, true);
            ConstructorInfo ci = ot.GetConstructor(new Type[0]);
            if (ci == null) //static
                o = null;
            else
                o = ci.Invoke(null);
        }

        public void CreateNew(string Type, object[] Args)
        {
            ot = System.Type.GetType(Type, true);
            Type[] types = new Type[Args.Length];
            for (int i = 0; i < Args.Length; i++)
                types[i] = Args[i].GetType();
            ConstructorInfo ci = ot.GetConstructor(types);
            o = ci.Invoke(Args);
        }

        public void FromControl(Control Control)
        {
            ot = Control.GetType();
            o = Control;
        }

        public void FromLibrary(string ObjectName, string FieldName, Hashtable htControls)
        {
            if (ObjectName.IndexOf(".") > -1)
                ObjectName = "_" + ObjectName.Replace(".", "_");
            else
                throw new Exception("ObjectName must start with the module name.\nExample: Main." +
                    ObjectName);
            ObjectName = ObjectName.ToLower();
            object o = htControls[ObjectName];
            if (o != null) //legacy
            {
                o = o.GetType().GetField("control").GetValue(o);
                o = o.GetType().GetField("Object").GetValue(o);
            }
            else //optimized compiler
            {
                o = htControls["_" + ObjectName];
            }
            o = o.GetType().GetField(FieldName, SysBindingFlags.Public | SysBindingFlags.Instance |
                    SysBindingFlags.Static | SysBindingFlags.NonPublic).GetValue(o);
            this.o = o;
            ot = o.GetType();
        }

        public void FromString(string name, Hashtable htCtls)
        {
            if (name.IndexOf(".") > -1)
                name = "_" + name.Replace(".", "_");
            else
                throw new Exception("Name must start with the module name.\nExample: Main." + name);
            name = name.ToLower();
            object o = htCtls[name];
            if (o != null) //legacy
            {
                o = o.GetType().GetField("control").GetValue(o);
                object o1 = o.GetType();
                string s = o1.ToString();
                if (o1.ToString() == "Dbasic.Other.SWrapObject")
                    o = o.GetType().GetField("Object").GetValue(o);
            }
            else //optimized compiler
            {
                o = htCtls["_" + name];
            }
            this.o = o;
            ot = this.o.GetType();
        }


        // exploratory methods
        private String mname(String method)
        {
            int i = method.IndexOf('(');
            if (i != -1)
                return method.Substring(0, i);
            else
                return method;
        }
        
        public string[] GetFields()
        {
            ArrayList al = new ArrayList();
            foreach (FieldInfo fi in ot.GetFields(bindingflags))
                al.Add(fi.Name + " As " + fi.FieldType.Name);
            al.Sort();
            return (string[])al.ToArray(typeof(string));
        }

        public string[] GetMethods()
        {
            ArrayList al = new ArrayList();
            foreach (MethodInfo mi in ot.GetMethods(bindingflags))
            {
                String n = mi.Name;
                ParameterInfo[] ps = mi.GetParameters();
                if (ps.Length > 0)
                    n = n + "(" + ps.Length + ")";
                if (n.Length > 4)
                {
                    if (n.Substring(0, 4) == "set_" || n.Substring(0, 4) == "get_" || n.Substring(0, 4) == "add_")
                    {
                        continue;
                    }
                }
                if (n.Length > 7 && n.Substring(0, 7) == "remove_")
                {
                    continue; ;
                }
                al.Add(n);
            }
            al.Sort();
            String last = "";
            ArrayList al2 = new ArrayList();
            foreach (String m in al)
            {
                if (mname(last) != mname(m))
                {
                    al2.Add(m); // method
                    last = m;
                }
                else // already added
                {
                    al2.RemoveAt(al2.Count - 1);
                    if (last.IndexOf('(') != -1)
                        last = last.Substring(0, last.Length - 1);
                    else
                        last = last + "(0";
                    last = last + ',' + m.Substring(m.Length - 2, 1) + ')';
                    al2.Add(last);
                }
            }
            return (string[])al2.ToArray(typeof(string));
        }

        public string[] GetMethodSignatures(string method)
        {
            ArrayList al = new ArrayList();
            MethodInfo[] mis = ot.GetMethods(bindingflags);
            if (mis != null)
            {
                foreach (MethodInfo mi in mis)
                {
                    if (mi.Name == method)
                    {
                        string m = mi.Name;
                        ParameterInfo[] pis = mi.GetParameters();
                        if (pis.Length > 0)
                        {
                            m += '(';
                            foreach (ParameterInfo pi in pis)
                            {
                                m += pi.Name + " As " + pi.ParameterType.Name + ", ";
                            }
                            m += ")";
                        }
                        al.Add(m + " As " + mi.ReturnType.Name);
                    }
                }
                al.Sort();
            }
            return (string[])al.ToArray(typeof(string));
        }

        public string[] GetProperties()
        {
            ArrayList al = new ArrayList();
            foreach (PropertyInfo pi in ot.GetProperties(bindingflags))
            {
                string rw = " [" + (pi.CanRead ? "r" : "") + (pi.CanWrite ? "w" : "");
                al.Add(pi.Name + " : " + pi.PropertyType.Name + ' ' + rw + ']');
            }
            al.Sort();
            return (string[])al.ToArray(typeof(string));
        }

        public string[] GetEvents()
        {
            ArrayList al = new ArrayList();
            foreach (EventInfo ei in o.GetType().GetEvents())
                al.Add(ei.Name + " : " + ei.EventHandlerType.Name);
            al.Sort();
            return (string[])al.ToArray(typeof(string));
        }

        public string[] GetEventArgsProperties(string eventname)
        {
            ArrayList al = new ArrayList();
            EventInfo ei = o.GetType().GetEvent(eventname);
            if (ei != null)
            {
                Type del = ei.EventHandlerType;
                MethodInfo invoke = del.GetMethod("Invoke");
                ParameterInfo[] pis = invoke.GetParameters();
                foreach (ParameterInfo pi in pis)
                {
                    if (pi.ParameterType.BaseType == typeof(EventArgs) ||
                        pi.ParameterType == typeof(EventArgs))
                    {
                        PropertyInfo[] peas = pi.ParameterType.GetProperties();
                        foreach (PropertyInfo pea in peas)
                            al.Add(pea.Name + " : " + pea.PropertyType.Name);
                    }
                }
                al.Sort();
                al.Insert(0, ei.EventHandlerType.Name + " properties");
            }
            return (string[])al.ToArray(typeof(string));
        }

        public void LoadAssembly(string PartialName)
        {
            #pragma warning disable 618
            o = Assembly.LoadWithPartialName(PartialName); // deprecated but useful
            #pragma warning restore 618
            ot = o.GetType();
        }

       
        /* The IDE can transfer data as object types between library objects.
         * It can only pass certain specific data from inside B4ppc to a property
         * or parameter of type Object such as Images, Arrays and B4PObjects.
         * No longer true. Can pass Strings
         * The optimising compiler can pass anything to an object type.
         */

        // GetField
        public object GetField(string Field)
        {
            FieldInfo fi = null;
            fi = ot.GetField(Field, bindingflags);
            lastField = fi;
            object ob = fi.GetValue(o);
            return ob;
        }

        public object GetFieldByInfo()
        {
            return lastField.GetValue(o);
        }

        // GetProperty
        public object GetProperty(string Property)
        {
            Type t = ot;
            PropertyInfo pi = null;
            try
            {
                pi = t.GetProperty(Property, bindingflags);
            }
            catch (AmbiguousMatchException)
            {
                foreach (PropertyInfo p in t.GetProperties(bindingflags))
                {
                    if (p.Name == Property)
                    {
                        pi = p;
                        break;
                    }
                }
            }
            lastProperty = pi;
            return pi.GetValue(o, null);
        }

        public object GetProperty2(string Property, int Index)
        {
            return GetProperty3(Property, (object)Index, "System.Int32");
        }

        public object GetProperty3(string Property, object Key, string KeyType)
        {
            Type keyt = Type.GetType(KeyType, true);
            PropertyInfo pi = null;
            try
            {
                pi = ot.GetProperty(Property, bindingflags);
            }
            catch (AmbiguousMatchException)
            {
                foreach (PropertyInfo p in ot.GetProperties(bindingflags))
                {
                    if (p.Name == Property)
                    {
                        if (p.GetIndexParameters()[0].ParameterType == keyt)
                        {
                            pi = p;
                            break;
                        }
                    }
                }
            }
            lastProperty = pi;
            object key = _Helper.StringToType(Key, keyt);
            return pi.GetValue(o, new object[] { key });
        }
        
        public object GetPropertyByInfo()
        {
            return lastProperty.GetValue(o, null);
        }
        
        public object GetPropertyByInfo2(object Key, string KeyType)
        {
            Type keyt = Type.GetType(KeyType, true);
            object key = _Helper.StringToType(Key, keyt);
            return lastProperty.GetValue(o, new object[] { key });
        }


        // RunMethod
        public object RunMethod(string Method)
        {
            return RunMethod4(Method, null);
        }

        public object RunMethod2(string Method, object Arg1, string Type1)
        {
            object[] oo = new object[1];
            oo[0] = _Helper.StringToType(Arg1, System.Type.GetType(Type1, true));
            return RunMethod4(Method, oo);
        }

        public object RunMethod3(string Method, object Arg1, string Type1, object Arg2, string Type2)
        {
            object[] oo = new object[2];
            oo[0] = _Helper.StringToType(Arg1, System.Type.GetType(Type1, true));
            oo[1] = _Helper.StringToType(Arg2, System.Type.GetType(Type2, true));
            return RunMethod4(Method, oo);
        }

        public object RunMethod4(string Method, object[] Args)
        {
            Type[] types;
            object ret;
            if (Args != null)
            {
                types = new Type[Args.Length];
                for (int i = 0; i < Args.Length; i++)
                    types[i] = Args[i].GetType();
                MethodInfo mi = ot.GetMethod(Method, bindingflags, null, types, null);
                lastMethod = mi;
                ret = mi.Invoke(o, Args);
            }
            else
            {
                MethodInfo mi = ot.GetMethod(Method, bindingflags, null, new Type[0], null);
                lastMethod = mi;
                ret = mi.Invoke(o, null);

            }
                return ret;
        }

        public object RunMethod5(string Method, object[] Args, string[] types)
        {
            Type[] typearr;
            object ret;
            if (Args != null)
            {
                typearr = new Type[Args.Length];
                for (int i = 0; i < Args.Length; i++)
                    typearr[i] = System.Type.GetType(types[i], true);
                MethodInfo mi = ot.GetMethod(Method, bindingflags, null, typearr, null);
                lastMethod = mi;
                ret = mi.Invoke(o, Args);
            }
            else
            {
                MethodInfo mi = ot.GetMethod(Method, bindingflags, null, new Type[0], null);
                lastMethod = mi;
                ret = mi.Invoke(o, null);

            }
            //if (ret == null)
            //    return "";
            //else
                return ret;
        }



        public object RunMethodByInfo(object[] Args)
        {
            return LastMethodInfo.Invoke(0, Args);
        }


        // SetField
        public void SetField(string Field, object Value)
        {
            FieldInfo fi = ot.GetField(Field, bindingflags);
            lastField = fi;
            Type ft = fi.FieldType;
            object y = _Helper.StringToType(Value, ft);
            fi.SetValue(o, y);
        }
        
        public void SetFieldByInfo(object Value)
        {
            Type ft = lastField.FieldType;
            object y = _Helper.StringToType(Value, ft);
            lastField.SetValue(o, y);
        }


        // SetProperty
        public void SetProperty(string Property, object Value)
        {
            Type t = ot;
            PropertyInfo pi = null;
            try
            {
                pi = t.GetProperty(Property, bindingflags);
            }
            catch (AmbiguousMatchException)
            {
                foreach (PropertyInfo p in t.GetProperties(bindingflags))
                {
                    if (p.Name == Property)
                    {
                        pi = p;
                        break;
                    }
                }
            }
            lastProperty = pi;
            Type v = pi.PropertyType;
            object y = _Helper.StringToType(Value, v);
            pi.SetValue(o, y, null);
        }

        public void SetProperty2(string Property, object Value, int Index)
        {
            SetProperty3(Property, Value, (object)Index, "System.Int32");
        }

        public void SetProperty3(string Property, object Value, object Key, string KeyType)
        {
            Type keyt = Type.GetType(KeyType, true);
            PropertyInfo pi = null;
            try
            {
                pi = ot.GetProperty(Property, bindingflags);

            }
            catch (AmbiguousMatchException)
            {
                foreach (PropertyInfo p in ot.GetProperties(bindingflags))
                {
                    if (p.Name == Property)
                    {
                        if (p.GetIndexParameters()[0].ParameterType == keyt)
                        {
                            pi = p;
                            break;
                        }
                    }
                }
            }
            lastProperty = pi;
            Type v = pi.PropertyType;
            object y = _Helper.StringToType(Value, v);
            object key = _Helper.StringToType(Key, keyt);
            pi.SetValue(o, y, new object[] { key });
        }

        public void SetPropertyByInfo(object Value)
        {
            Type v = lastProperty.PropertyType;
            object y = _Helper.StringToType(Value, v);
            lastProperty.SetValue(o, y, null);
        }

        public void SetPropertyByInfo2(object Value, object Key, string KeyType)
        {
            Type keyt = Type.GetType(KeyType, true);
            Type v = lastProperty.PropertyType;
            object y = _Helper.StringToType(Value, v);
            object key = _Helper.StringToType(Key, keyt);
            lastProperty.SetValue(o, y, new object[] { key });
        }



        // Public properties

        public int BindingFlags
        { get { return (int) bindingflags; } set { bindingflags = (System.Reflection.BindingFlags)value;}}

        public double DLLVersion
        {
            get { return version; }
        }

        public int FlagDeclaredOnly
        { get { return (int)SysBindingFlags.DeclaredOnly; } }

        public int FlagIgnoreCase
        { get { return (int)SysBindingFlags.IgnoreCase; } }

        public int FlagInstance
        { get { return (int)SysBindingFlags.Instance; } }

        public int FlagNonPublic
        { get { return (int)SysBindingFlags.NonPublic; } }

        public int FlagPublic
        { get { return (int)SysBindingFlags.Public; } }

        public int FlagStatic
        { get { return (int)SysBindingFlags.Static; } }


        public FieldInfo LastFieldinfo
        { get { return lastField; } set { lastField = value; } }

        public MethodInfo LastMethodInfo
        { get { return lastMethod; } set { lastMethod = value; } }

        public PropertyInfo LastPropertyInfo
        { get { return lastProperty; } set { lastProperty = value; } }

        public object Null
        { get { return null; } }

        public object Value
        {
            get { return o; }
            set { o = value; ot = (o == null ? null : o.GetType()); }
        }

        public string ValueAsEnum
        { get { return _Helper.EnumToString(o); } set { o = Enum.Parse(ot, value, true); } }

        public bool ValueAsBoolean
        { get { return o.ToString().ToLower() == "true" ? true : false; } set { o = value; ot = o.GetType(); } }

        public char ValueAsChar
        { get { return Convert.ToChar(o); } set { o = value; ot = o.GetType(); } }

        public int ValueAsColor
        { get { return ((Color)o).ToArgb(); } set { o = Color.FromArgb(value); ot = o.GetType(); } }

        public double ValueAsDouble
        { get { return Convert.ToDouble(o); } set { o = value; ot = o.GetType(); } }

        public Single ValueAsSingle
        { get { return Convert.ToSingle(o); } set { o = value; ot = o.GetType(); } }

        public short ValueAsInt16
        { get { return Convert.ToInt16(o); } set { o = value; ot = o.GetType(); } }

        public int ValueAsInt32
        { get { return Convert.ToInt32(o); } set { o = value; ot = o.GetType(); } }

        public long ValueAsInt64
        { get { return Convert.ToInt64(o); } set { o = value; ot = o.GetType(); } }

        public IntPtr ValueAsIntPtr
        { get { return (IntPtr)o; } set { o = value; ot = o.GetType(); } }

        public string ValueAsString
        { get { return o.ToString(); } set { o = value; ot = o.GetType(); } }

        public ushort ValueAsUInt16
        { get { return Convert.ToUInt16(o); } set { o = value; ot = o.GetType(); } }

        public uint ValueAsUInt32
        { get { return Convert.ToUInt32(o); } set { o = value; ot = o.GetType(); } }

        public ulong ValueAsUInt64
        { get { return Convert.ToUInt64(o); } set { o = value; ot = o.GetType(); } }

        public bool ValueIsNull
        { get { return o == null; } }

        public string ValueToString
        { get {
            if (o == null)
                return "null";
            return o.ToString();
        } }
 
        public string ValueType()
        {
            if (o == null)
                if (ot == null)
                    return "null";
                else
                    return ot.ToString() + " Type";
            return ot.ToString();
        }

        public string ValueTypeName()
        {
            if (o == null)
                if (ot == null)
                    return "null";
                else
                    return ot.Name + " Type";
            return ot.Name;
        }



    }

    public class ObjectArray
    {
        private object[] o;

        // constructor
        public ObjectArray(int Size)
        {
            o = new object[Size];
        }

        // public methods
        public object Get(int Index)
        {
            return o[Index];
        }

        public void SetObject(int Index, object Object)
        {
            o[Index] = Object;
        }

        public void SetValue(int Index, string Value, string Type)
        {
            o[Index] = _Helper.StringToType(Value, System.Type.GetType(Type, true));
        }

        // public property
        public object[] Value
        { get { return o; } set { o = value; }  }

    }

    // original Event object code with HelpEventHandler type added
    public class Event : IDisposable
    {
        private object[] eventObject;
        public event EventHandler NewEvent;
        private object args = null;

        // constructor
        public Event(object Object, string Event)
        {
            eventObject = new object[] { this, "NewEvent" };

            //EventInfo ei = Object.GetType().GetEvent(Event, BindingFlags.IgnoreCase | BindingFlags.Public
            //   | BindingFlags.Static | BindingFlags.Instance); 
            //System.Collections.Generic.Dictionary<Type, Delegate> d = new System.Collections.Generic.Dictionary<Type, Delegate>(16);
            //d.Add(typeof(System.ComponentModel.CancelEventHandler), new System.ComponentModel.CancelEventHandler(eventHandler));
            //d.Add(typeof(EventHandler), new EventHandler(eventHandler));
            //d.Add(typeof(KeyEventHandler), new KeyEventHandler(eventHandler));
            //d.Add(typeof(KeyPressEventHandler), new KeyPressEventHandler(eventHandler));
            //d.Add(typeof(MouseEventHandler), new MouseEventHandler(eventHandler));
            //d.Add(typeof(PaintEventHandler), new PaintEventHandler(eventHandler));
            //d.Add(typeof(HelpEventHandler), new HelpEventHandler(eventHandler));
            //ei.AddEventHandler(Object, d[ei.EventHandlerType]);

            // Get an EventInfo object representing the event, 
            EventInfo ei = Object.GetType().GetEvent(Event, BindingFlags.IgnoreCase | BindingFlags.Public
               | BindingFlags.Static | BindingFlags.Instance); 

           // Use the EventHandlerType to get the type of delegate used to handle the event. 
            Type eht = ei.EventHandlerType;
            
            // Get a MethodInfo object representing the method that handles the event
            MethodInfo mi = typeof(Event).GetMethod("eventHandler", BindingFlags.NonPublic | BindingFlags.Instance);
            
            // Create an instance of the delegate, using the CreateDelegate method.
            Delegate ehd = Delegate.CreateDelegate(eht, this, mi);

            // Get the add accessor method and invoke it to hook up the event.
            // All events have an add accessor and a remove accessor, which are hidden by the syntax of high-level languages.
            // For example, C# uses the += operator 
            MethodInfo addmi = ei.GetAddMethod();
            addmi.Invoke(Object, new object[] { ehd } );
        }

        // private event handler
        private void eventHandler(object sender, object e)
        {
            args = e;
            if (NewEvent != null)
                NewEvent(eventObject, null);
        }

        // public property
        public object Data
        { get { return args; } }


        public void Dispose()
        {
            NewEvent = null;
        }
    }

    // original _Helper code with null handling added to both methods
    public static class _Helper
    {
        static System.Globalization.CultureInfo cul = new CultureInfo("en-US", false);
        
        public static string EnumToString(object o)
        {
            // added to original
            if (o == null)
                return "";
            // end added
            Type t = o.GetType();
            if (t.IsEnum)
            {
                return ((Enum)o).ToString();
            }
            return o.ToString();
        }

        public static object StringToType(object o, Type t)
        {
            // added to original
            if (o == null)
                return o;
            // end added
            if (o.GetType() == typeof(string))
            {
                if (t == typeof(string))
                    return o;
                else if (t == typeof(Char))
                {
                    return ((string)o)[0];
                }
                else if (t.IsEnum)
                {
                    return Enum.Parse(t, (string)o, true);
                }
                else if (t == typeof(bool))
                {
                    return bool.Parse(((string)o).ToLower(cul));
                }

                if (t.IsPrimitive)
                {
                    double d = double.Parse((string)o, cul);
                    return Convert.ChangeType(d, t, cul);
                }
            }
            return o;
        }

    }

}
